///////////////////////////////////////////////////////////////////////
///
/// \file   SignalShapingServiceT962.h
///
/// \brief  Service to provide microboone-specific signal shaping for
///         simulation (convolution) and reconstruction (deconvolution).
///
/// \author H. Greenlee, B. Baller ported to ArgoNeuT
///
/// This service inherits from SignalShaping and supplies
/// ArgoNeuT-specific configuration.  It is intended that SimWire and
/// CalWire modules will access this service.
///
/// FCL parameters:
///
/// FieldBins       - Number of bins of field response.
/// Col3DCorrection - 3D path length correction for collection plane.
/// Ind3DCorrection - 3D path length correction for induction plane.
/// ColFieldRespAmp - Collection field response amplitude.
/// IndFieldRespAmp - Induction field response amplitude.
/// IndFieldParams  - Induction field function parameters
/// ShapeTimeConst  - Time constants for exponential shaping.
/// ColFilter       - Root parameterized collection plane filter function.
/// ColFilterParams - Collection filter function parameters.
/// IndFilter       - Root parameterized induction plane filter function.
/// IndFilterParams - Induction filter function parameters.
///
////////////////////////////////////////////////////////////////////////

#ifndef SIGNALSHAPINGSERVICET962_H
#define SIGNALSHAPINGSERVICET962_H

#include <vector>
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Services/Registry/ActivityRegistry.h"
#include "art/Framework/Services/Registry/ServiceMacros.h"
#include "lardata/Utilities/SignalShaping.h"
#include "TF1.h"
#include "TH1D.h"

using DoubleVec = std::vector<double>;

namespace util {
  class SignalShapingServiceT962 {
  public:

    // Constructor, destructor.

    SignalShapingServiceT962(const fhicl::ParameterSet& pset,
				   art::ActivityRegistry& reg);
    ~SignalShapingServiceT962();

    // Update configuration parameters.

    void reconfigure(const fhicl::ParameterSet& pset);

//    std::vector<DoubleVec> GetNoiseFactVec()                { return fNoiseFactVec; }

//    double GetRawNoise(unsigned int const channel) const ;
//    double GetDeconNoise(unsigned int const channel) const;   
    
    // Accessors.

    const util::SignalShaping& SignalShaping(unsigned int channel) const;

//    int FieldResponseTOffset(unsigned int const channel) const;

    // Do convolution calcution (for simulation).

    template <class T> void Convolute(unsigned int channel, std::vector<T>& func) const;

    // Do deconvolution calcution (for reconstruction).

    template <class T> void Deconvolute(unsigned int channel, std::vector<T>& func) const;
    
    double GetDeconNorm(){return fDeconNorm;};

  private:

    // Private configuration methods.

    const static unsigned short nChans = 480;
    unsigned short fftSize;
    
    // Post-constructor initialization.

    void init() const{const_cast<SignalShapingServiceT962*>(this)->init();}
    void init();

    // Stored response, e.g. electronics and readout cable dependent long-tail exponential
    std::string  fResponseFile;      ///< response file containing 

    void SetFieldResponse();
    void SetElectResponse();

    // Calculate filter functions.

    void SetFilters();

    // Attributes.

    bool fInit;               ///< Initialization flag.

    // Fcl parameters.

    std::vector<float> fIndFieldParams;         ///< induction plane field function parameterization
    std::vector<float> fColFieldParams;         ///< collection plane field function parameterization

    double fDeconNorm;
    TF1* fColFilterFunc;      			            ///< Parameterized collection filter function.
    TF1* fIndFilterFunc;      			            ///< Parameterized induction filter function.

    // response per TDC tick
    std::vector<double> fElectResponse;
    // exponential tail time constants for each group of channels. Normally only two kernels would
    // be required (induction and collection) but ArgoNeuT has long expontial tails that varies
    // in groups of channels that depend on the cable. This vector holds the exponential tail time
    // constants for each group.
    double fLongTailGroupTolerance;
    std::vector<double> fTailConsts;
    std::array<unsigned short, nChans> fChanGroup;
    // convolution kernels for each group of channels
    std::vector<util::SignalShaping> fSignalShapeGroup;

    // Field response.
    std::vector<double> fColFieldResponse;
    std::vector<double> fIndFieldResponse;

    std::vector<TComplex> fColFilter;
    std::vector<TComplex> fIndFilter;
  };
}
//----------------------------------------------------------------------
// Do convolution.
template <class T> inline void util::SignalShapingServiceT962::Convolute(unsigned int channel, std::vector<T>& func) const
{
  SignalShaping(channel).Convolute(func);

  //negative number
  int time_offset = 0;
//  int time_offset = FieldResponseTOffset(channel);
  
  std::vector<T> temp;
  if (time_offset <=0){
    temp.assign(func.begin(),func.begin()-time_offset);
    func.erase(func.begin(),func.begin()-time_offset);
    func.insert(func.end(),temp.begin(),temp.end());
  }else{
    temp.assign(func.end()-time_offset,func.end());
    func.erase(func.end()-time_offset,func.end());
    func.insert(func.begin(),temp.begin(),temp.end());
  }
}


//----------------------------------------------------------------------
// Do deconvolution.
template <class T> inline void util::SignalShapingServiceT962::Deconvolute(unsigned int channel, std::vector<T>& func) const
{
  SignalShaping(channel).Deconvolute(func);

  int time_offset = 0;
//   int time_offset = FieldResponseTOffset(channel);
  
  std::vector<T> temp;
  if (time_offset <=0){
    temp.assign(func.end()+time_offset,func.end());
    func.erase(func.end()+time_offset,func.end());
    func.insert(func.begin(),temp.begin(),temp.end());
  }else{
    temp.assign(func.begin(),func.begin()+time_offset);
    func.erase(func.begin(),func.begin()+time_offset);
    func.insert(func.end(),temp.begin(),temp.end());    
  }
}

DECLARE_ART_SERVICE(util::SignalShapingServiceT962, LEGACY)
#endif
